import time

import pytest
from aioxmpp import JID

from spade import quit_spade
from spade.container import Container
from spade.message import Message


@pytest.fixture
def jid():
    return JID.fromstr("friend@localhost/home")


@pytest.fixture
def message():
    return Message(
        to="to@localhost",
        sender="sender@localhost",
        body="message body",
        thread="thread-id",
        metadata={"metadata1": "value1", "metadata2": "value2"},
    )


@pytest.fixture
def message2():
    return Message(
        to="to2@localhost",
        sender="sender2@localhost",
        body="message body",
        thread="thread-id",
    )


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    # A test function will be run at this point
    container = Container()
    if not container.is_running:
        container.__init__()
    yield
    # Code that will run after your test, for example:
    quit_spade()


@pytest.fixture(scope="module", autouse=True)
def cleanup(request):
    quit_spade()


def wait_for_behaviour_is_killed(behaviour, tries=500, sleep=0.01):
    counter = 0
    while not behaviour.is_killed() and counter < tries:
        time.sleep(sleep)
        counter += 1
    if not behaviour.is_killed():
        raise Exception("Behaviour not finished")



import asyncio
import time

import requests
from aiohttp import web
from aiohttp_jinja2 import get_env
from aioxmpp import JID, PresenceType
from aioxmpp.roster import Item
from asynctest import Mock, CoroutineMock, MagicMock
from jinja2 import ChoiceLoader, FileSystemLoader, PackageLoader
from parsel import Selector
from testfixtures import LogCapture

from spade.agent import Agent
from spade.behaviour import OneShotBehaviour, CyclicBehaviour
from spade.message import Message
from .factories import MockedAgentFactory, MockedPresenceAgentFactory


def test_web():
    agent = MockedAgentFactory()
    future = agent.start(auto_register=False)
    future.result()
    agent.web.start(port=10000)

    assert agent.web.app is not None

    counter = 0
    while counter < 4:
        if agent.web.server is not None:
            break
        counter += 1
        time.sleep(0.1)
    assert agent.web.server is not None
    agent.stop()


def test_default_template_path():
    agent = Agent("jid@server", "password")

    env = get_env(agent.web.app)
    loader = env.loader

    package_loader = loader.loaders[0]
    filesystem_loader = loader.loaders[1]

    assert type(loader) == ChoiceLoader
    assert len(loader.loaders) == 2
    assert type(package_loader) == PackageLoader
    assert type(filesystem_loader) == FileSystemLoader

    assert "internal_tpl_agent.html" in package_loader.list_templates()
    assert "internal_tpl_agent.html" not in filesystem_loader.list_templates()
    assert filesystem_loader.searchpath == ["."]

    agent.stop()


def test_add_template_path():
    agent = Agent("jid@server", "password")

    agent.web.start(templates_path="/tmp/spade")

    env = get_env(agent.web.app)
    loader = env.loader

    assert type(loader) == ChoiceLoader
    assert len(loader.loaders) == 3
    assert type(loader.loaders[0]) == FileSystemLoader
    assert type(loader.loaders[1]) == PackageLoader
    assert type(loader.loaders[2]) == FileSystemLoader

    filesystem_loader = loader.loaders[0]

    assert filesystem_loader.list_templates() == []
    assert filesystem_loader.searchpath == ["/tmp/spade"]

    agent.stop()


async def test_check_server(test_client):
    agent = MockedAgentFactory()
    future = agent.start(auto_register=False)
    future.result()

    agent.web.setup_routes()

    client = await test_client(agent.web.app)
    response = await client.get("/spade")

    text = await response.text()

    sel = Selector(text=text)

    assert sel.css("title::text").get() == "fake agent"
    assert sel.css("img::attr(src)").get() == agent.avatar

    assert sel.css("ul.products-list > li").getall() == []

    agent.stop().result()


async def test_request_home(test_client):
    agent = MockedAgentFactory(jid="jid@server", password="password")
    future = agent.start(auto_register=False)
    future.result()
    agent.web.setup_routes()
    client = await test_client(agent.web.app)

    response = await client.get("/spade")
    response = await response.text()

    sel = Selector(text=response)

    assert sel.css("title::text").get() == "jid agent"
    assert sel.css("img::attr(src)").get() == agent.avatar

    assert sel.css("ul.products-list > li").getall() == []

    agent.stop().result()


async def test_get_messages(test_client):
    agent = Agent("jid@server", "password")
    agent.web.setup_routes()
    client = await test_client(agent.web.app)

    # add messages to trace
    for i in range(5):
        msg = Message(body=str(i), sender="{}@server".format(i), to="receiver@server")
        agent.traces.append(msg)

    response = await client.get("/spade/messages/")
    response = await response.text()

    sel = Selector(text=response)

    assert len(sel.css("ul.timeline > li").getall()) == 6  # num messages + end clock

    agent.stop().result()


async def test_get_behaviour(test_client):
    class EmptyOneShotBehaviour(OneShotBehaviour):
        async def run(self):
            self.kill()

    agent = Agent("jid@server", "password")
    behaviour = EmptyOneShotBehaviour()
    agent.add_behaviour(behaviour)
    agent.web.setup_routes()

    client = await test_client(agent.web.app)

    response = await client.get(
        "/spade/behaviour/OneShotBehaviour/EmptyOneShotBehaviour/"
    )
    response = await response.text()

    sel = Selector(text=response)

    assert (
        sel.css("section.content-header > h1::text").get().strip()
        == "OneShotBehaviour/EmptyOneShotBehaviour"
    )
    agent.stop().result()


async def test_kill_behaviour(test_client):
    class EmptyCyclicBehaviour(CyclicBehaviour):
        async def run(self):
            await asyncio.sleep(0.1)

    agent = Agent("jid@server", "password")
    behaviour = EmptyCyclicBehaviour()
    agent.add_behaviour(behaviour)

    agent.web.setup_routes()
    client = await test_client(agent.web.app)

    await client.get("/spade/behaviour/CyclicBehaviour/EmptyCyclicBehaviour/kill/")

    assert behaviour.is_killed()

    agent.stop().result()


async def test_get_agent(test_client):
    agent = MockedPresenceAgentFactory(jid="jid@server", password="password")
    future = agent.start(auto_register=False)
    future.result()

    agent.web.setup_routes()
    client = await test_client(agent.web.app)

    jid = "friend@server"
    item = Item(jid=JID.fromstr(jid))

    agent.presence.roster._update_entry(item)

    response = await client.get(f"/spade/agent/{jid}/")
    response = await response.text()

    sel = Selector(text=response)

    assert sel.css("section.content-header > h1::text").get().strip() == jid

    agent.stop().result()


async def test_unsubscribe_agent(test_client):
    agent = MockedPresenceAgentFactory()
    future = agent.start(auto_register=False)
    future.result()
    agent.client.enqueue = Mock()

    agent.web.setup_routes()
    client = await test_client(agent.web.app)

    jid = "friend@server"
    jid_ = JID.fromstr(jid)
    item = Item(jid=jid_)

    agent.presence.roster._update_entry(item)

    response = await client.get(f"/spade/agent/{jid}/unsubscribe/")

    assert str(response.url.relative()) == f"/spade/agent/{jid}/"

    assert agent.client.enqueue.mock_calls
    arg = agent.client.enqueue.call_args[0][0]

    assert arg.to == jid_.bare()
    assert arg.type_ == PresenceType.UNSUBSCRIBE

    agent.stop().result()


async def test_send_agent(test_client):
    agent = MockedPresenceAgentFactory()
    future = agent.start(auto_register=False)
    future.result()
    agent.stream = MagicMock()
    agent.stream.send = CoroutineMock()
    agent.web.setup_routes()
    client = await test_client(agent.web.app)

    jid = "friend@server"
    item = Item(jid=JID.fromstr(jid))
    agent.presence.roster._update_entry(item)

    msg = "Hello World"

    response = await client.post(f"/spade/agent/{jid}/send/", data={"message": msg})

    assert str(response.url.relative()) == f"/spade/agent/{jid}/"

    sent = agent.traces.all()[0]

    assert sent[1].sent
    assert sent[1].body == "Hello World"

    agent.stop().result()


async def test_find_behaviour():
    class EmptyOneShotBehaviour(OneShotBehaviour):
        async def run(self):
            pass

    agent = MockedAgentFactory()
    behaviour = EmptyOneShotBehaviour()
    agent.add_behaviour(behaviour)
    found_behaviour = agent.web.find_behaviour("OneShotBehaviour/EmptyOneShotBehaviour")

    assert found_behaviour == behaviour

    agent.stop().result()


async def test_find_behaviour_fail():
    agent = Agent("jid@server", "password")
    found_behaviour = agent.web.find_behaviour("OneShotBehaviour/EmptyOneShotBehaviour")

    assert found_behaviour is None

    agent.stop().result()


async def test_add_get(test_client):
    agent = Agent("jid@server", "password")
    agent.web.add_get("/test", lambda request: {"number": 42}, "tests/hello.html")

    agent.web.setup_routes()
    client = await test_client(agent.web.app)

    response = await client.get("/test")
    response = await response.text()

    sel = Selector(text=response)
    assert sel.css("h1::text").get().strip() == "42"

    agent.stop().result()


async def test_add_get_raw(test_client):
    agent = Agent("jid@server", "password")
    agent.web.add_get(
        "/test",
        lambda request: web.Response(text="Hello Raw Get"),
        template=None,
        raw=True,
    )

    agent.web.setup_routes()
    client = await test_client(agent.web.app)

    response = await client.get("/test")
    response = await response.text()

    assert response == "Hello Raw Get"

    agent.stop().result()


async def test_add_post(test_client):
    agent = Agent("jid@server", "password")

    async def handle_post(request):
        form = await request.post()
        number = form["number"]
        return {"number": number}

    agent.web.add_post("/test", handle_post, "tests/hello.html")
    agent.web.setup_routes()
    client = await test_client(agent.web.app)

    response = await client.post("/test", data={"number": 1024})
    response = await response.text()

    sel = Selector(text=response)
    assert sel.css("h1::text").get() == "1024"

    agent.stop().result()


async def test_add_post_raw(test_client):
    agent = Agent("jid@server", "password")

    async def handle_post(request):
        form = await request.post()
        number = form["number"]
        return web.Response(text="Hello Raw Post Number={}".format(number))

    agent.web.add_post("/test", handle_post, template=None, raw=True)
    agent.web.setup_routes()
    client = await test_client(agent.web.app)

    response = await client.post("/test", data={"number": 1024})
    response = await response.text()

    assert response == "Hello Raw Post Number=1024"

    agent.stop().result()


async def test_stop(test_client):
    agent = Agent("jid@server", "password")
    agent.web.setup_routes()
    client = await test_client(agent.web.app)

    response = await client.get("/spade/stop")
    response = await response.text()

    sel = Selector(text=response)
    assert (
        sel.css("div.alert-warning > span::text").get().strip()
        == "Agent is stopping now."
    )

    with LogCapture() as log:
        try:
            await client.get("/spade/stop/now/", timeout=0.05)
        except requests.exceptions.ReadTimeout:
            pass
        except asyncio.exceptions.TimeoutError:
            pass

        log.check_present(
            ("spade.Web", "WARNING", "Stopping agent from web interface.")
        )

    counter = 5
    while agent.is_alive() and counter > 0:
        counter -= 0.5
        time.sleep(0.5)

    assert not agent.is_alive()


async def test_add_get_json(test_client):
    async def controller(request):
        return {"number": 42}

    agent = Agent("jid@server", "password")
    agent.web.add_get("/test", controller, None)

    agent.web.setup_routes()
    client = await test_client(agent.web.app)

    response = await client.get("/test")
    assert response.status == 200

    data = await response.json()
    assert data["number"] == 42

    agent.stop().result()


async def test_add_post_json(test_client):
    async def handle_post(request):
        form = await request.post()
        number = form["number"]
        return {"number": int(number)}

    agent = Agent("jid@server", "password")
    agent.web.add_post("/test", handle_post, None)

    agent.web.setup_routes()
    client = await test_client(agent.web.app)

    response = await client.post("/test", data={"number": 1024})
    assert response.status == 200

    data = await response.json()
    assert data["number"] == 1024

    agent.stop().result()



from unittest.mock import Mock

import pytest
from aioxmpp import PresenceState, PresenceShow, JID, PresenceType, Presence
from aioxmpp.roster.xso import Item as XSOItem

from spade.presence import ContactNotFound
from .factories import MockedPresenceAgentFactory


def test_get_state_not_available():
    agent = MockedPresenceAgentFactory(available=False, show=PresenceShow.NONE)

    future = agent.start(auto_register=False)
    future.result()

    assert type(agent.presence.state) == PresenceState
    assert agent.presence.state.available is False
    assert agent.presence.state.show == PresenceShow.NONE
    assert not agent.presence.is_available()


def test_get_state_available():
    agent = MockedPresenceAgentFactory(available=True)

    future = agent.start(auto_register=False)
    future.result()
    agent.mock_presence()

    assert agent.presence.state.available
    assert agent.presence.is_available()


def test_set_available():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.set_available()

    assert agent.presence.is_available()


def test_set_available_with_show():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.set_available(show=PresenceShow.CHAT)

    assert agent.presence.is_available()
    assert agent.presence.state.show == PresenceShow.CHAT


def test_set_unavailable():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.set_unavailable()

    assert not agent.presence.is_available()


def test_get_state_show():
    agent = MockedPresenceAgentFactory(available=True, show=PresenceShow.AWAY)

    future = agent.start(auto_register=False)
    future.result()
    agent.mock_presence()

    assert agent.presence.state.show == PresenceShow.AWAY


def test_get_status_empty():
    agent = MockedPresenceAgentFactory(status={})

    future = agent.start(auto_register=False)
    future.result()
    agent.mock_presence()

    assert agent.presence.status == {}


def test_get_status_string():
    agent = MockedPresenceAgentFactory(status="Working")

    future = agent.start(auto_register=False)
    future.result()
    agent.mock_presence()

    assert agent.presence.status == {None: "Working"}


def test_get_status_dict():
    agent = MockedPresenceAgentFactory(status={"en": "Working"})

    future = agent.start(auto_register=False)
    future.result()
    agent.mock_presence()

    assert agent.presence.status == {"en": "Working"}


def test_get_priority_default():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()
    agent.mock_presence()

    assert agent.presence.priority == 0


def test_get_priority():
    agent = MockedPresenceAgentFactory(priority=10)

    future = agent.start(auto_register=False)
    future.result()
    agent.mock_presence()

    assert agent.presence.priority == 10


def test_set_presence_available():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.set_presence(state=PresenceState(available=True))

    assert agent.presence.is_available()


def test_set_presence_unavailable():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.set_presence(state=PresenceState(available=False))

    assert not agent.presence.is_available()


def test_set_presence_status():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.set_presence(status="Lunch")

    assert agent.presence.status == {None: "Lunch"}


def test_set_presence_status_dict():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.set_presence(status={"en": "Lunch"})

    assert agent.presence.status == {"en": "Lunch"}


def test_set_presence_priority():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.set_presence(priority=5)

    assert agent.presence.priority == 5


def test_set_presence():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.set_presence(
        state=PresenceState(True, PresenceShow.PLAIN), status="Lunch", priority=2
    )

    assert agent.presence.is_available()
    assert agent.presence.state.show == PresenceShow.PLAIN
    assert agent.presence.status == {None: "Lunch"}
    assert agent.presence.priority == 2


def test_get_contacts_empty():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    assert agent.presence.get_contacts() == {}


def test_get_contacts(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    item = XSOItem(jid=jid)
    item.approved = True
    item.name = "My Friend"

    agent.presence.roster._update_entry(item)

    contacts = agent.presence.get_contacts()

    bare_jid = jid.bare()
    assert bare_jid in contacts
    assert type(contacts[bare_jid]) == dict
    assert contacts[bare_jid]["approved"]
    assert contacts[bare_jid]["name"] == "My Friend"
    assert contacts[bare_jid]["subscription"] == "none"
    assert "ask" not in contacts[bare_jid]
    assert "groups" not in contacts[bare_jid]


def test_get_contacts_with_presence(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    item = XSOItem(jid=jid)
    item.approved = True
    item.name = "My Available Friend"

    agent.presence.roster._update_entry(item)

    stanza = Presence(from_=jid, type_=PresenceType.AVAILABLE)
    agent.presence.presenceclient.handle_presence(stanza)

    contacts = agent.presence.get_contacts()

    bare_jid = jid.bare()
    assert bare_jid in contacts
    assert contacts[bare_jid]["name"] == "My Available Friend"

    assert contacts[bare_jid]["presence"].type_ == PresenceType.AVAILABLE


def test_get_contacts_with_presence_on_and_off(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    item = XSOItem(jid=jid)
    item.approved = True
    item.name = "My Friend"

    agent.presence.roster._update_entry(item)

    stanza = Presence(from_=jid, type_=PresenceType.AVAILABLE)
    agent.presence.presenceclient.handle_presence(stanza)
    stanza = Presence(from_=jid, type_=PresenceType.UNAVAILABLE)
    agent.presence.presenceclient.handle_presence(stanza)

    contacts = agent.presence.get_contacts()

    bare_jid = jid.bare()
    assert bare_jid in contacts
    assert contacts[bare_jid]["name"] == "My Friend"

    assert contacts[bare_jid]["presence"].type_ == PresenceType.UNAVAILABLE


def test_get_contacts_with_presence_unavailable(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    item = XSOItem(jid=jid)
    item.approved = True
    item.name = "My UnAvailable Friend"

    agent.presence.roster._update_entry(item)

    stanza = Presence(from_=jid, type_=PresenceType.UNAVAILABLE)
    agent.presence.presenceclient.handle_presence(stanza)

    contacts = agent.presence.get_contacts()

    bare_jid = jid.bare()
    assert bare_jid in contacts
    assert contacts[bare_jid]["name"] == "My UnAvailable Friend"

    assert "presence" not in contacts[bare_jid]


def test_get_contact(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    item = XSOItem(jid=jid)
    item.approved = True
    item.name = "My Friend"

    agent.presence.roster._update_entry(item)

    contact = agent.presence.get_contact(jid)

    assert type(contact) == dict
    assert contact["approved"]
    assert contact["name"] == "My Friend"
    assert contact["subscription"] == "none"
    assert "ask" not in contact
    assert "groups" not in contact


def test_get_invalid_jid_contact():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    with pytest.raises(ContactNotFound):
        agent.presence.get_contact(JID.fromstr("invalid@contact"))


def test_get_invalid_str_contact():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    with pytest.raises(AttributeError):
        agent.presence.get_contact("invalid@contact")


def test_subscribe(jid):
    peer_jid = str(jid)
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.client.enqueue = Mock()
    agent.presence.subscribe(peer_jid)

    assert agent.client.enqueue.mock_calls
    arg = agent.client.enqueue.call_args[0][0]

    assert arg.to == jid.bare()
    assert arg.type_ == PresenceType.SUBSCRIBE


def test_unsubscribe(jid):
    peer_jid = str(jid)
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.client.enqueue = Mock()
    agent.presence.unsubscribe(peer_jid)

    assert agent.client.enqueue.mock_calls
    arg = agent.client.enqueue.call_args[0][0]

    assert arg.to == jid.bare()
    assert arg.type_ == PresenceType.UNSUBSCRIBE


def test_approve(jid):
    peer_jid = str(jid)
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.client.enqueue = Mock()
    agent.presence.approve(peer_jid)

    assert agent.client.enqueue.mock_calls
    arg = agent.client.enqueue.call_args[0][0]

    assert arg.to == jid.bare()
    assert arg.type_ == PresenceType.SUBSCRIBED


def test_on_available(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.on_available = Mock()

    stanza = Presence(from_=jid, type_=PresenceType.AVAILABLE)
    agent.presence.presenceclient.handle_presence(stanza)

    jid_arg = agent.presence.on_available.call_args[0][0]
    stanza_arg = agent.presence.on_available.call_args[0][1]

    assert jid_arg == str(jid)
    assert stanza_arg.type_ == PresenceType.AVAILABLE


def test_on_unavailable(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.on_unavailable = Mock()
    agent.presence.presenceclient._presences[jid.bare()] = {"home": None}

    stanza = Presence(from_=jid, type_=PresenceType.UNAVAILABLE)
    agent.presence.presenceclient.handle_presence(stanza)

    jid_arg = agent.presence.on_unavailable.call_args[0][0]
    stanza_arg = agent.presence.on_unavailable.call_args[0][1]

    assert jid_arg == str(jid)
    assert stanza_arg.type_ == PresenceType.UNAVAILABLE


def test_on_subscribe(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.on_subscribe = Mock()

    stanza = Presence(from_=jid, type_=PresenceType.SUBSCRIBE)
    agent.presence.roster.handle_subscribe(stanza)

    jid_arg = agent.presence.on_subscribe.call_args[0][0]

    assert jid_arg == str(jid)


def test_on_subscribe_approve_all(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.approve_all = True
    agent.client.enqueue = Mock()

    stanza = Presence(from_=jid, type_=PresenceType.SUBSCRIBE)
    agent.presence.roster.handle_subscribe(stanza)

    assert agent.client.enqueue.mock_calls
    arg = agent.client.enqueue.call_args[0][0]

    assert arg.to == jid.bare()
    assert arg.type_ == PresenceType.SUBSCRIBED


def test_on_subscribed(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.on_subscribed = Mock()

    stanza = Presence(from_=jid, type_=PresenceType.SUBSCRIBED)
    agent.presence.roster.handle_subscribed(stanza)

    jid_arg = agent.presence.on_subscribed.call_args[0][0]

    assert jid_arg == str(jid)


def test_on_unsubscribe(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.on_unsubscribe = Mock()

    stanza = Presence(from_=jid, type_=PresenceType.UNSUBSCRIBE)
    agent.presence.roster.handle_unsubscribe(stanza)

    jid_arg = agent.presence.on_unsubscribe.call_args[0][0]

    assert jid_arg == str(jid)


def test_on_unsubscribe_approve_all(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.approve_all = True
    agent.client.enqueue = Mock()

    stanza = Presence(from_=jid, type_=PresenceType.UNSUBSCRIBE)
    agent.presence.roster.handle_unsubscribe(stanza)

    assert agent.client.enqueue.mock_calls
    arg = agent.client.enqueue.call_args[0][0]

    assert arg.to == jid.bare()
    assert arg.type_ == PresenceType.UNSUBSCRIBED


def test_on_unsubscribed(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    agent.presence.on_unsubscribed = Mock()

    stanza = Presence(from_=jid, type_=PresenceType.UNSUBSCRIBED)
    agent.presence.roster.handle_unsubscribed(stanza)

    jid_arg = agent.presence.on_unsubscribed.call_args[0][0]

    assert jid_arg == str(jid)


def test_on_changed(jid):
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    item = XSOItem(jid=jid)
    item.approved = True
    item.name = "My Friend"

    agent.presence.roster._update_entry(item)

    stanza = Presence(from_=jid, type_=PresenceType.AVAILABLE, show=PresenceShow.CHAT)
    agent.presence.presenceclient.handle_presence(stanza)

    contact = agent.presence.get_contact(jid)
    assert contact["name"] == "My Friend"
    assert contact["presence"].show == PresenceShow.CHAT

    stanza = Presence(from_=jid, type_=PresenceType.AVAILABLE, show=PresenceShow.AWAY)
    agent.presence.presenceclient.handle_presence(stanza)

    contact = agent.presence.get_contact(jid)

    assert contact["name"] == "My Friend"
    assert contact["presence"].show == PresenceShow.AWAY


def test_ignore_self_presence():
    agent = MockedPresenceAgentFactory()

    future = agent.start(auto_register=False)
    future.result()

    jid = agent.jid

    stanza = Presence(from_=jid, type_=PresenceType.AVAILABLE, show=PresenceShow.CHAT)
    agent.presence.presenceclient.handle_presence(stanza)

    with pytest.raises(ContactNotFound):
        agent.presence.get_contact(jid)

    assert len(agent.presence.get_contacts()) == 0



# -*- coding: utf-8 -*-

from . import agent
from . import behaviour
from . import message
from . import template
from .container import stop_container as quit_spade  # noqa: F401

__author__ = """Javi Palanca"""
__email__ = "jpalanca@gmail.com"
__version__ = "3.2.0"

__all__ = ["agent", "behaviour", "message", "template"]



import asyncio
import logging
import sys
from hashlib import md5
from threading import Event

import aiosasl
import aioxmpp
import aioxmpp.ibr as ibr
from aioxmpp.dispatcher import SimpleMessageDispatcher

from spade.behaviour import FSMBehaviour
from spade.container import Container
from spade.message import Message
from spade.presence import PresenceManager
from spade.trace import TraceStore
from spade.web import WebApp

logger = logging.getLogger("spade.Agent")


class AuthenticationFailure(Exception):
    """ """

    pass


class Agent(object):
    def __init__(self, jid, password, verify_security=False):
        """
        Creates an agent

        Args:
          jid (str): The identifier of the agent in the form username@server
          password (str): The password to connect to the server
          verify_security (bool): Wether to verify or not the SSL certificates
        """
        self.jid = aioxmpp.JID.fromstr(jid)
        self.password = password
        self.verify_security = verify_security

        self.behaviours = []
        self._values = {}

        self.conn_coro = None
        self.stream = None
        self.client = None
        self.message_dispatcher = None
        self.presence = None
        self.loop = None

        self.container = Container()
        self.container.register(self)

        self.loop = self.container.loop

        # Web service
        self.web = WebApp(agent=self)

        self.traces = TraceStore(size=1000)

        self._alive = Event()

    def set_loop(self, loop):
        self.loop = loop

    def set_container(self, container):
        """
        Sets the container to which the agent is attached

        Args:
            container (spade.container.Container): the container to be attached to
        """
        self.container = container

    def start(self, auto_register=True):
        """
        Tells the container to start this agent.
        It returns a coroutine or a future depending on whether it is called from a coroutine or a synchronous method.

        Args:
            auto_register (bool): register the agent in the server (Default value = True)
        """
        return self.container.start_agent(agent=self, auto_register=auto_register)

    async def _async_start(self, auto_register=True):
        """
        Starts the agent from a coroutine. This fires some actions:

            * if auto_register: register the agent in the server
            * runs the event loop
            * connects the agent to the server
            * runs the registered behaviours

        Args:
          auto_register (bool, optional): register the agent in the server (Default value = True)

        """

        await self._hook_plugin_before_connection()

        if auto_register:
            await self._async_register()
        self.client = aioxmpp.PresenceManagedClient(
            self.jid,
            aioxmpp.make_security_layer(
                self.password, no_verify=not self.verify_security
            ),
            loop=self.loop,
            logger=logging.getLogger(self.jid.localpart),
        )

        # obtain an instance of the service
        self.message_dispatcher = self.client.summon(SimpleMessageDispatcher)

        # Presence service
        self.presence = PresenceManager(self)

        await self._async_connect()

        # register a message callback here
        self.message_dispatcher.register_callback(
            aioxmpp.MessageType.CHAT, None, self._message_received,
        )

        await self._hook_plugin_after_connection()

        await self.setup()
        self._alive.set()
        for behaviour in self.behaviours:
            if not behaviour.is_running:
                behaviour.set_agent(self)
                if issubclass(type(behaviour), FSMBehaviour):
                    for _, state in behaviour.get_states().items():
                        state.set_agent(self)
                behaviour.start()

    async def _hook_plugin_before_connection(self):
        """
        Overload this method to hook a plugin before connetion is done
        """
        pass

    async def _hook_plugin_after_connection(self):
        """
        Overload this method to hook a plugin after connetion is done
        """
        pass

    async def _async_connect(self):  # pragma: no cover
        """ connect and authenticate to the XMPP server. Async mode. """
        try:
            self.conn_coro = self.client.connected()
            aenter = type(self.conn_coro).__aenter__(self.conn_coro)
            self.stream = await aenter
            logger.info(f"Agent {str(self.jid)} connected and authenticated.")
        except aiosasl.AuthenticationFailure:
            raise AuthenticationFailure(
                "Could not authenticate the agent. Check user and password or use auto_register=True"
            )

    async def _async_register(self):  # pragma: no cover
        """ Register the agent in the XMPP server from a coroutine. """
        metadata = aioxmpp.make_security_layer(None, no_verify=not self.verify_security)
        query = ibr.Query(self.jid.localpart, self.password)
        _, stream, features = await aioxmpp.node.connect_xmlstream(
            self.jid, metadata, loop=self.loop
        )
        await ibr.register(stream, query)

    async def setup(self):
        """
        Setup agent before startup.
        This coroutine may be overloaded.
        """
        await asyncio.sleep(0)

    @property
    def name(self):
        """ Returns the name of the agent (the string before the '@') """
        return self.jid.localpart

    @property
    def avatar(self):
        """
        Generates a unique avatar for the agent based on its JID.
        Uses Gravatar service with MonsterID option.

        Returns:
          str: the url of the agent's avatar

        """
        return self.build_avatar_url(self.jid.bare())

    @staticmethod
    def build_avatar_url(jid):
        """
        Static method to build a gravatar url with the agent's JID

        Args:
          jid (aioxmpp.JID): an XMPP identifier

        Returns:
          str: an URL for the gravatar

        """
        digest = md5(str(jid).encode("utf-8")).hexdigest()
        return "http://www.gravatar.com/avatar/{md5}?d=monsterid".format(md5=digest)

    def submit(self, coro):
        """
        Runs a coroutine in the event loop of the agent.
        this call is not blocking.

        Args:
          coro (coroutine): the coroutine to be run

        Returns:
            asyncio.Future: the future of the coroutine execution

        """
        return asyncio.run_coroutine_threadsafe(coro, loop=self.loop)

    def add_behaviour(self, behaviour, template=None):
        """
        Adds and starts a behaviour to the agent.
        If template is not None it is used to match
        new messages and deliver them to the behaviour.

        Args:
          behaviour (spade.behaviour.CyclicBehaviour): the behaviour to be started
          template (spade.template.Template, optional): the template to match messages with (Default value = None)

        """
        behaviour.set_agent(self)
        if issubclass(type(behaviour), FSMBehaviour):
            for _, state in behaviour.get_states().items():
                state.set_agent(self)
        behaviour.set_template(template)
        self.behaviours.append(behaviour)
        if self.is_alive():
            behaviour.start()

    def remove_behaviour(self, behaviour):
        """
        Removes a behaviour from the agent.
        The behaviour is first killed.

        Args:
          behaviour (spade.behaviour.CyclicBehaviour): the behaviour instance to be removed

        """
        if not self.has_behaviour(behaviour):
            raise ValueError("This behaviour is not registered")
        index = self.behaviours.index(behaviour)
        self.behaviours[index].kill()
        self.behaviours.pop(index)

    def has_behaviour(self, behaviour):
        """
        Checks if a behaviour is added to an agent.

        Args:
          behaviour (spade.behaviour.CyclicBehaviour): the behaviour instance to check

        Returns:
          bool: a boolean that indicates wether the behaviour is inside the agent.

        """
        return behaviour in self.behaviours

    def stop(self):
        """
        Tells the container to start this agent.
        It returns a coroutine or a future depending on whether it is called from a coroutine or a synchronous method.
        """
        return self.container.stop_agent(self)

    async def _async_stop(self):
        """ Stops an agent and kills all its behaviours. """
        if self.presence:
            self.presence.set_unavailable()
        for behav in self.behaviours:
            behav.kill()
        if self.web.is_started():
            await self.web.runner.cleanup()

        """ Discconnect from XMPP server. """
        if self.is_alive():
            # Disconnect from XMPP server
            self.client.stop()
            aexit = self.conn_coro.__aexit__(*sys.exc_info())
            await aexit
            logger.info("Client disconnected.")

        self._alive.clear()

    def is_alive(self):
        """
        Checks if the agent is alive.

        Returns:
          bool: wheter the agent is alive or not

        """
        return self._alive.is_set()

    def set(self, name, value):
        """
        Stores a knowledge item in the agent knowledge base.

        Args:
          name (str): name of the item
          value (object): value of the item

        """
        self._values[name] = value

    def get(self, name):
        """
        Recovers a knowledge item from the agent's knowledge base.

        Args:
          name(str): name of the item

        Returns:
          object: the object retrieved or None

        """
        if name in self._values:
            return self._values[name]
        else:
            return None

    def _message_received(self, msg):
        """
        Callback run when an XMPP Message is reveived.
        This callback delivers the message to every behaviour
        that is waiting for it. First, the aioxmpp.Message is
        converted to spade.message.Message

        Args:
          msg (aioxmpp.Messagge): the message just received.

        Returns:
            list(asyncio.Future): a list of futures of the append of the message at each matched behaviour.

        """

        msg = Message.from_node(msg)
        return self.dispatch(msg)

    def dispatch(self, msg):
        """
        Dispatch the message to every behaviour that is waiting for
        it using their templates match.

        Args:
          msg (spade.message.Messagge): the message to dispatch.

        Returns:
            list(asyncio.Future): a list of futures of the append of the message at each matched behaviour.

        """
        logger.debug(f"Got message: {msg}")
        futures = []
        matched = False
        for behaviour in (x for x in self.behaviours if x.match(msg)):
            futures.append(self.submit(behaviour.enqueue(msg)))
            logger.debug(f"Message enqueued to behaviour: {behaviour}")
            self.traces.append(msg, category=str(behaviour))
            matched = True
        if not matched:
            logger.warning(f"No behaviour matched for message: {msg}")
            self.traces.append(msg)
        return futures



import asyncio

import aioxmpp
from aioxmpp import PresenceManagedClient
from asynctest import CoroutineMock, Mock
from testfixtures import LogCapture

from spade.agent import Agent
from spade.behaviour import OneShotBehaviour
from spade.message import Message
from spade.template import Template
from .factories import MockedAgentFactory


def test_create_agent(mocker):
    agent = Agent("jid@server", "fake_password")
    agent._async_connect = CoroutineMock()

    assert agent.is_alive() is False

    future = agent.start(auto_register=False)
    assert future.result() is None

    agent._async_connect.assert_called_once()
    assert agent.stream is None

    agent.conn_coro = mocker.Mock()
    agent.conn_coro.__aexit__ = CoroutineMock()

    assert agent.is_alive() is True
    future = agent.stop()
    future.result()

    agent.conn_coro.__aexit__.assert_called_once()

    assert agent.is_alive() is False


def test_connected_agent():
    agent = MockedAgentFactory()
    assert agent.is_alive() is False

    future = agent.start(auto_register=False)
    assert future.result() is None
    assert agent.is_alive() is True

    future = agent.stop()
    future.result()
    assert agent.is_alive() is False


def test_name():
    agent = MockedAgentFactory(jid="john@fake_server")
    assert agent.name == "john"


def test_avatar():
    agent = MockedAgentFactory(jid="test_avatar@fake_server")
    assert (
        agent.avatar
        == "http://www.gravatar.com/avatar/44bdc5585ef57844edb11c5b9711d2e6?d=monsterid"
    )


def test_setup():
    agent = MockedAgentFactory()
    agent.setup = CoroutineMock()
    future = agent.start(auto_register=False)
    assert future.result() is None

    agent.setup.assert_called_once()
    agent.stop()


def test_set_get():
    agent = MockedAgentFactory()
    agent.set("KB_name", "KB_value")
    assert agent.get("KB_name") == "KB_value"


def test_get_none():
    agent = MockedAgentFactory()
    assert agent.get("KB_name_unknown") is None


def test_client():
    agent = MockedAgentFactory()
    assert agent.client is None

    future = agent.start()
    future.result()
    assert type(agent.client) == PresenceManagedClient


def test_register():
    agent = MockedAgentFactory()
    agent.register = Mock()

    future = agent.start(auto_register=True)
    assert future.result() is None

    assert len(agent._async_register.mock_calls) == 1

    agent.stop()


def test_receive_without_behaviours():
    agent = MockedAgentFactory()
    aiomsg = aioxmpp.Message(type_=aioxmpp.MessageType.CHAT)
    msg = Message.from_node(aiomsg)

    assert agent.traces.len() == 0
    future = agent.start(auto_register=False)
    assert future.result() is None

    with LogCapture() as log:
        agent._message_received(aiomsg)
        log.check_present(
            ("spade.Agent", "WARNING", f"No behaviour matched for message: {msg}")
        )

    assert agent.traces.len() == 1
    assert msg in agent.traces.store[0]

    agent.stop()


def test_create_agent_from_another_agent():
    class DummyBehav(OneShotBehaviour):
        async def run(self):
            self.agent._done = True
            self.kill()

    class CreateBehav(OneShotBehaviour):
        async def run(self):
            self.agent.agent2 = MockedAgentFactory()
            self.agent.agent2._done = False
            self.agent.dummy_behav = DummyBehav()
            self.agent.agent2.add_behaviour(self.agent.dummy_behav)
            await self.agent.agent2.start(auto_register=False)
            self.kill()

    agent1 = MockedAgentFactory()
    agent1.agent2 = None
    create_behav = CreateBehav()
    agent1.add_behaviour(create_behav)
    future = agent1.start(auto_register=False)
    assert future.result() is None
    assert agent1.is_alive()

    create_behav.join()
    agent1.dummy_behav.join()

    assert agent1.agent2.is_alive()
    assert agent1.agent2._done

    agent1.agent2.stop()
    agent1.stop()


def test_create_agent_from_another_agent_from_setup():
    class DummyBehav(OneShotBehaviour):
        async def run(self):
            self.agent._done = True
            self.kill()

    class SetupAgent(Agent):
        async def setup(self):
            self.agent2 = MockedAgentFactory()
            self.agent2._done = False
            self.agent2.dummy_behav = DummyBehav()
            self.agent2.add_behaviour(self.agent2.dummy_behav)
            await self.agent2.start(auto_register=False)

    agent1 = SetupAgent("fake@host", "secret")
    agent1._async_connect = CoroutineMock()
    agent1._async_register = CoroutineMock()
    agent1.conn_coro = Mock()
    agent1.conn_coro.__aexit__ = CoroutineMock()
    agent1.stream = Mock()

    agent1.agent2 = None

    future = agent1.start(auto_register=False)
    assert future.result() is None
    assert agent1.is_alive()

    agent1.agent2.dummy_behav.join()

    assert agent1.agent2.is_alive()
    assert agent1.agent2._done

    agent1.agent2.stop()
    agent1.stop()


def test_submit_send():
    agent = MockedAgentFactory()

    class DummyBehav(OneShotBehaviour):
        async def run(self):
            self.agent.recv_msg = await self.receive(10)

    template = Template(to="fake@jid")
    behav = DummyBehav()
    agent.add_behaviour(behav, template=template)

    future = agent.start(auto_register=False)
    future.result()

    msg_to_send = Message(to="fake@jid", body="BODY", metadata={"performative": "TEST"})
    agent.submit(behav.send(msg_to_send))
    behav.join()

    assert str(agent.recv_msg.to) == "fake@jid"
    assert agent.recv_msg.body == "BODY"
    assert agent.recv_msg.metadata == {"performative": "TEST"}


def test_stop_agent_with_blocking_await():
    agent1 = MockedAgentFactory()
    agent1.value = 1000

    class StopBehav(OneShotBehaviour):
        async def run(self):
            await asyncio.sleep(0.5)
            await self.agent.stop()

    class DummyBehav(OneShotBehaviour):
        async def run(self):
            await self.receive(timeout=1000000)
            self.agent.value = 2000

    stopbehah = StopBehav()
    dummybehav = DummyBehav()

    agent1.add_behaviour(dummybehav)
    agent1.add_behaviour(stopbehah)

    future1 = agent1.start(auto_register=False)
    future1.result()

    stopbehah.join()

    assert not agent1.is_alive()
    assert agent1.value == 1000



import asyncio
import time

import aioxmpp
from asynctest import MagicMock, CoroutineMock

from spade.behaviour import OneShotBehaviour, CyclicBehaviour
from spade.container import Container
from spade.message import Message
from .factories import MockedAgentFactory


def test_use_container():
    container = Container()
    container.reset()

    agent = MockedAgentFactory()

    assert agent.container == Container()

    assert container.has_agent(str(agent.jid))
    assert container.get_agent(str(agent.jid)) == agent

    agent.stop()


def test_send_message_with_container():
    class FakeReceiverAgent:
        def __init__(self):
            self.jid = "fake_receiver_agent@server"

        def set_container(self, c):
            pass

        def set_loop(self, loop):
            pass

        def stop(self):
            pass

        def is_alive(self):
            return False

    class SendBehaviour(OneShotBehaviour):
        async def run(self):
            message = Message(to="fake_receiver_agent@server")
            await self.send(message)
            self.kill()

    container = Container()
    container.reset()
    fake_receiver_agent = FakeReceiverAgent()
    container.register(fake_receiver_agent)

    fake_receiver_agent.dispatch = MagicMock()

    agent = MockedAgentFactory()
    future = agent.start(auto_register=False)
    future.result()

    agent.client = MagicMock()
    agent.client.send = CoroutineMock()
    behaviour = SendBehaviour()
    agent.add_behaviour(behaviour)

    behaviour.join()

    assert agent.client.send.await_count == 0

    assert fake_receiver_agent.dispatch.call_count == 1
    assert (
        str(fake_receiver_agent.dispatch.call_args[0][0].to)
        == "fake_receiver_agent@server"
    )

    agent.stop()


def test_send_message_to_outer_with_container():
    class SendBehaviour(OneShotBehaviour):
        async def run(self):
            message = Message(to="to@outerhost")
            await self.send(message)
            self.kill()

    container = Container()
    container.reset()

    agent = MockedAgentFactory()
    agent.start(auto_register=False)

    behaviour = SendBehaviour()
    behaviour._xmpp_send = CoroutineMock()
    agent.add_behaviour(behaviour)

    behaviour.join()

    assert container.has_agent(str(agent.jid))
    assert not container.has_agent("to@outerhost")

    assert behaviour._xmpp_send.await_count == 1
    msg_arg = behaviour._xmpp_send.await_args[0][0]
    assert msg_arg.to == aioxmpp.JID.fromstr("to@outerhost")

    agent.stop()


def test_unregister():
    container = Container()
    container.reset()

    agent = MockedAgentFactory()
    agent2 = MockedAgentFactory(jid="agent2@server")

    assert container.has_agent(str(agent.jid))
    assert container.has_agent(str(agent2.jid))

    container.unregister(agent.jid)

    assert not container.has_agent(str(agent.jid))
    assert container.has_agent(str(agent2.jid))


def test_cancel_tasks():
    agent = MockedAgentFactory()

    class Behav(CyclicBehaviour):
        async def run(self):
            await asyncio.sleep(100)
            self.has_finished = True

    behav = Behav()
    behav.has_finished = False
    agent.add_behaviour(behaviour=behav)
    future = agent.start()
    future.result()

    assert not behav.has_finished

    container = Container()
    container.stop()

    assert not behav.has_finished


def test_stop_container():
    agent = MockedAgentFactory()

    class Behav(OneShotBehaviour):
        async def run(self):
            container = Container()
            container.stop()

    behav = Behav()
    agent.add_behaviour(behav)
    future = agent.start(auto_register=False)
    future.result()
    behav.join()

    counter = 0
    while agent.is_alive() and counter < 5:
        time.sleep(0.05)
        counter += 1
    assert not agent.is_alive()



import asyncio
import logging
import sys
from contextlib import suppress
from threading import Thread

from singletonify import singleton

logger = logging.getLogger("SPADE")


@singleton()
class Container(object):
    """
     The container class allows agents to exchange messages
     without using the XMPP socket if they are in the same
     process.
     The container is a singleton.
    """

    def __init__(self):
        self.__agents = {}
        self.aiothread = AioThread()
        self.aiothread.start()
        self.loop = self.aiothread.loop
        self.loop.set_debug(False)
        self.is_running = True

    def __in_coroutine(self):
        try:
            return asyncio.get_event_loop() == self.loop
        except RuntimeError:  # pragma: no cover
            return False

    def start_agent(self, agent, auto_register=True):
        coro = agent._async_start(auto_register=auto_register)

        if self.__in_coroutine():
            return coro
        else:
            return asyncio.run_coroutine_threadsafe(coro, loop=self.loop)

    def stop_agent(self, agent):
        coro = agent._async_stop()

        if self.__in_coroutine():
            return coro
        else:
            return asyncio.run_coroutine_threadsafe(coro, loop=self.loop)

    def reset(self):
        """ Empty the container by unregistering all the agents. """
        self.__agents = {}

    def register(self, agent):
        """
        Register a new agent.

        Args:
            agent (spade.agent.Agent): the agent to be registered
        """
        self.__agents[str(agent.jid)] = agent
        agent.set_container(self)
        agent.set_loop(self.loop)

    def unregister(self, jid):
        if str(jid) in self.__agents:
            del self.__agents[str(jid)]

    def has_agent(self, jid):
        """
        Check if an agent is registered in the container.
        Args:
            jid (str): the jid of the agent to be checked.

        Returns:
            bool: wether the agent is or is not registered.
        """
        return jid in self.__agents

    def get_agent(self, jid):
        """
        Returns a registered agent
        Args:
            jid (str): the identifier of the agent

        Returns:
            spade.agent.Agent: the agent you were looking for

        Raises:
            KeyError: if the agent is not found
        """
        return self.__agents[jid]

    async def send(self, msg, behaviour):
        """
        This method sends the message using the container mechanism
        when the receiver is also registered in the container. Otherwise,
        it uses the XMPP send method from the original behaviour.
        Args:
            msg (spade.message.Message): the message to be sent
            behaviour: the behaviour that is sending the message
        """
        to = str(msg.to)
        if to in self.__agents:
            self.__agents[to].dispatch(msg)
        else:
            await behaviour._xmpp_send(msg)

    def stop(self):
        agents = [agent.stop() for agent in self.__agents.values() if agent.is_alive()]
        if self.__in_coroutine():
            coro = asyncio.gather(*agents)
            asyncio.run_coroutine_threadsafe(coro, loop=self.loop)
        self.aiothread.finalize()
        self.reset()
        self.is_running = False


class AioThread(Thread):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loop = asyncio.new_event_loop()
        self.running = True
        self.daemon = True

    def run(self):
        try:
            self.loop.run_forever()
            if sys.version_info >= (3, 7):
                tasks = asyncio.all_tasks(loop=self.loop)  # pragma: no cover
            else:
                tasks = asyncio.Task.all_tasks(loop=self.loop)  # pragma: no cover
            for task in tasks:
                task.cancel()
                with suppress(asyncio.CancelledError):
                    self.loop.run_until_complete(task)
            self.loop.close()
            logger.debug("Loop closed")
        except Exception as e:  # pragma: no cover
            logger.error("Exception in the event loop: {}".format(e))

    def finalize(self):
        if self.running:
            self.loop.call_soon_threadsafe(self.loop.stop)
            self.running = False


def stop_container():
    container = Container()
    container.stop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `message` package."""
import copy

import aioxmpp
import aioxmpp.forms.xso as forms_xso
import pytest

from spade.message import Message, SPADE_X_METADATA


def test_prepare(message):
    aiomsg = message.prepare()
    assert aiomsg.to == aioxmpp.JID.fromstr("to@localhost")
    assert aiomsg.from_ == aioxmpp.JID.fromstr("sender@localhost")
    assert aiomsg.body[None] == "message body"

    for data in aiomsg.xep0004_data:
        if data.title == SPADE_X_METADATA:
            for field in data.fields:
                if field.var == "_thread_node":
                    assert field.values[0] == "thread-id"
                else:
                    assert message.get_metadata(field.var) == field.values[0]


def test_make_reply(message):
    reply = message.make_reply()
    assert reply.to == aioxmpp.JID.fromstr("sender@localhost")
    assert reply.sender == aioxmpp.JID.fromstr("to@localhost")
    assert reply.body == "message body"
    assert reply.thread == "thread-id"
    assert reply.get_metadata("metadata1") == "value1"
    assert reply.get_metadata("metadata2") == "value2"


def test_message_from_node_attribute_error():
    with pytest.raises(AttributeError) as e:
        Message.from_node(Message())


def test_body_with_languages():
    msg = aioxmpp.Message(type_=aioxmpp.MessageType.CHAT)
    msg.body["en"] = "Hello World"
    msg.body["es"] = "Hola Mundo"

    new_msg = Message.from_node(msg)
    assert new_msg.body == "Hello World"


def test_message_from_node():
    aiomsg = aioxmpp.Message(type_=aioxmpp.MessageType.CHAT)
    data = forms_xso.Data(type_=forms_xso.DataType.FORM)

    data.fields.append(
        forms_xso.Field(
            var="performative",
            type_=forms_xso.FieldType.TEXT_SINGLE,
            values=["request"],
        )
    )

    data.fields.append(
        forms_xso.Field(
            var="_thread_node",
            type_=forms_xso.FieldType.TEXT_SINGLE,
            values=["thread-id"],
        )
    )
    data.title = SPADE_X_METADATA
    aiomsg.xep0004_data = [data]

    msg = Message.from_node(aiomsg)

    assert msg.thread == "thread-id"
    assert msg.get_metadata("performative") == "request"
    assert msg.metadata == {"performative": "request"}


def test_thread_empty():
    msg = Message(thread=None)

    assert msg.thread is None
    assert msg.metadata == {}

    aiomsg = msg.prepare()
    for data in aiomsg.xep0004_data:
        if data.title == SPADE_X_METADATA:
            for field in data.fields:
                assert field.var != "_thread_node"


def test_equal(message):
    assert message == copy.copy(message)


def test_not_equal(message, message2):
    assert message != message2


def test_id(message):
    assert type(message.id) == int


def test_metadata_is_string():
    Message(metadata={"key": "value"})


def test_metadata_is_not_string():
    with pytest.raises(TypeError):
        Message(metadata={"key": 1000})


def test_metadata_set_string():
    msg = Message()
    msg.set_metadata("key", "value")


def test_metadata_set_not_string():
    msg = Message()
    with pytest.raises(TypeError):
        msg.set_metadata(1000, "value")


def test_body_is_string():
    Message(body="body")


def test_body_is_not_string():
    with pytest.raises(TypeError):
        Message(body={})


def test_body_set_string():
    msg = Message()
    msg.body = "body"


def test_body_set_not_string():
    msg = Message()
    with pytest.raises(TypeError):
        msg.body = 1000


def test_body_set_none():
    msg = Message()
    msg.body = None


def test_to_is_string():
    Message(to="agent@fakeserver")


def test_to_is_not_string():
    with pytest.raises(TypeError):
        Message(to=aioxmpp.JID.fromstr("agent@fakeserver"))


def test_to_set_string():
    msg = Message()
    msg.to = "agent@fakeserver"


def test_to_set_not_string():
    msg = Message()
    with pytest.raises(TypeError):
        msg.to = 1000


def test_to_set_none():
    msg = Message()
    msg.to = None


def test_sender_is_string():
    Message(sender="agent@fakeserver")


def test_sender_is_not_string():
    with pytest.raises(TypeError):
        Message(sender=aioxmpp.JID.fromstr("agent@fakeserver"))


def test_sender_set_string():
    msg = Message()
    msg.sender = "agent@fakeserver"


def test_sender_set_not_string():
    msg = Message()
    with pytest.raises(TypeError):
        msg.sender = 1000


def test_sender_set_none():
    msg = Message()
    msg.sender = None


def test_thread_is_string():
    Message(thread="thread_id_001")


def test_thread_is_not_string():
    with pytest.raises(TypeError):
        Message(thread=1000)


def test_thread_set_string():
    msg = Message()
    msg.thread = "thread_id_001"


def test_thread_set_not_string():
    msg = Message()
    with pytest.raises(TypeError):
        msg.thread = 1000


def test_thread_set_none():
    msg = Message()
    msg.thread = None



import aioxmpp
from aioxmpp import PresenceState, PresenceShow


class ContactNotFound(Exception):
    """ """

    pass


class PresenceManager(object):
    """ """

    def __init__(self, agent):
        self.agent = agent
        self.client = agent.client
        self.roster = self.client.summon(aioxmpp.RosterClient)
        self.presenceclient = self.client.summon(aioxmpp.PresenceClient)
        self.presenceserver = self.client.summon(aioxmpp.PresenceServer)

        self._contacts = {}

        self.approve_all = False

        self.presenceclient.on_bare_available.connect(self._on_bare_available)
        self.presenceclient.on_available.connect(self._on_available)
        self.presenceclient.on_bare_unavailable.connect(self._on_bare_unavailable)
        self.presenceclient.on_unavailable.connect(self._on_unavailable)
        self.presenceclient.on_changed.connect(self._on_changed)

        self.roster.on_subscribe.connect(self._on_subscribe)
        self.roster.on_subscribed.connect(self._on_subscribed)
        self.roster.on_unsubscribe.connect(self._on_unsubscribe)
        self.roster.on_unsubscribed.connect(self._on_unsubscribed)

    @property
    def state(self):
        """
        The currently set presence state (as aioxmpp.PresenceState)
        which is broadcast when the client connects and when the presence is
        re-emitted.

        This attribute cannot be written. It does not reflect the actual
        presence seen by others. For example when the client is in fact
        offline, others will see unavailable presence no matter what is set
        here.

        Returns:
            aioxmpp.PresenceState: the presence state of the agent
        """
        return self.presenceserver.state

    @property
    def status(self):
        """
        The currently set textual presence status which is broadcast when the
        client connects and when the presence is re-emitted.

        This attribute cannot be written. It does not reflect the actual
        presence seen by others. For example when the client is in fact
        offline, others will see unavailable presence no matter what is set
        here.

        Returns:
            dict: a dict with the status in different languages (default key is None)
        """
        return self.presenceserver.status

    @property
    def priority(self):
        """
        The currently set priority which is broadcast when the client connects
        and when the presence is re-emitted.

        This attribute cannot be written. It does not reflect the actual
        presence seen by others. For example when the client is in fact
        offline, others will see unavailable presence no matter what is set
        here.

        Returns:
            int: the priority of the connection
        """
        return self.presenceserver.priority

    def is_available(self):
        """
        Returns the available flag from the state

        Returns:
          bool: wether the agent is available or not

        """
        return self.state.available

    def set_available(self, show=PresenceShow.NONE):
        """
        Sets the agent availability to True.

        Args:
          show (aioxmpp.PresenceShow, optional): the show state of the presence (Default value = PresenceShow.NONE)

        """
        show = self.state.show if show is PresenceShow.NONE else show
        self.set_presence(PresenceState(available=True, show=show))

    def set_unavailable(self):
        """Sets the agent availability to False."""
        show = PresenceShow.NONE
        self.set_presence(PresenceState(available=False, show=show))

    def set_presence(self, state=None, status=None, priority=None):
        """
        Change the presence broadcast by the client.
        If the client is currently connected, the new presence is broadcast immediately.

        Args:
          state(aioxmpp.PresenceState, optional): New presence state to broadcast (Default value = None)
          status(dict or str, optional): New status information to broadcast (Default value = None)
          priority (int, optional): New priority for the resource (Default value = None)

        """
        state = state if state is not None else self.state
        status = status if status is not None else self.status
        priority = priority if priority is not None else self.priority
        self.presenceserver.set_presence(state, status, priority)

    def get_contacts(self):
        """
        Returns list of contacts

        Returns:
          dict: the roster of contacts

        """
        for jid, item in self.roster.items.items():
            try:
                self._contacts[jid.bare()].update(item.export_as_json())
            except KeyError:
                self._contacts[jid.bare()] = item.export_as_json()

        return self._contacts

    def get_contact(self, jid):
        """
        Returns a contact

        Args:
          jid (aioxmpp.JID): jid of the contact

        Returns:
          dict: the roster of contacts

        """
        try:
            return self.get_contacts()[jid.bare()]
        except KeyError:
            raise ContactNotFound
        except AttributeError:
            raise AttributeError("jid must be an aioxmpp.JID object")

    def _update_roster_with_presence(self, stanza):
        """ """
        if stanza.from_.bare() == self.agent.jid.bare():
            return
        try:
            self._contacts[stanza.from_.bare()].update({"presence": stanza})
        except KeyError:
            self._contacts[stanza.from_.bare()] = {"presence": stanza}

    def subscribe(self, peer_jid):
        """
        Asks for subscription

        Args:
          peer_jid (str): the JID you ask for subscriptiion

        """
        self.roster.subscribe(aioxmpp.JID.fromstr(peer_jid).bare())

    def unsubscribe(self, peer_jid):
        """
        Asks for unsubscription

        Args:
          peer_jid (str): the JID you ask for unsubscriptiion

        """
        self.roster.unsubscribe(aioxmpp.JID.fromstr(peer_jid).bare())

    def approve(self, peer_jid):
        """
        Approve a subscription request from jid

        Args:
          peer_jid (str): the JID to approve

        """
        self.roster.approve(aioxmpp.JID.fromstr(peer_jid).bare())

    def _on_bare_available(self, stanza):
        """ """
        self._update_roster_with_presence(stanza)
        self.on_available(str(stanza.from_), stanza)

    def _on_available(self, full_jid, stanza):
        """ """
        self._update_roster_with_presence(stanza)
        self.on_available(str(stanza.from_), stanza)

    def _on_unavailable(self, full_jid, stanza):
        """ """
        self._update_roster_with_presence(stanza)
        self.on_unavailable(str(stanza.from_), stanza)

    def _on_bare_unavailable(self, stanza):
        """ """
        self._update_roster_with_presence(stanza)
        self.on_unavailable(str(stanza.from_), stanza)

    def _on_changed(self, from_, stanza):
        """ """
        self._update_roster_with_presence(stanza)

    def _on_subscribe(self, stanza):
        """ """
        if self.approve_all:
            self.roster.approve(stanza.from_.bare())
        else:
            self.on_subscribe(str(stanza.from_))

    def _on_subscribed(self, stanza):
        """ """
        self.on_subscribed(str(stanza.from_))

    def _on_unsubscribe(self, stanza):
        """ """
        if self.approve_all:
            self.client.stream.enqueue(
                aioxmpp.Presence(
                    type_=aioxmpp.structs.PresenceType.UNSUBSCRIBED,
                    to=stanza.from_.bare(),
                )
            )
        else:
            self.on_unsubscribe(str(stanza.from_))

    def _on_unsubscribed(self, stanza):
        """ """
        self.on_unsubscribed(str(stanza.from_))

    def on_subscribe(self, peer_jid):
        """
        Callback called when a subscribe query is received.
        To ve overloaded by user.

        Args:
          peer_jid (str): the JID of the agent asking for subscription

        """
        pass  # pragma: no cover

    def on_subscribed(self, peer_jid):
        """
        Callback called when a subscribed message is received.
        To ve overloaded by user.

        Args:
          peer_jid (str): the JID of the agent that accepted subscription

        """
        pass  # pragma: no cover

    def on_unsubscribe(self, peer_jid):
        """
        Callback called when an unsubscribe query is received.
        To ve overloaded by user.

        Args:
          peer_jid (str): the JID of the agent asking for unsubscription

        """
        pass  # pragma: no cover

    def on_unsubscribed(self, peer_jid):
        """
        Callback called when an unsubscribed message is received.
        To ve overloaded by user.

        Args:
          peer_jid (str): the JID of the agent that unsubscribed

        """
        pass  # pragma: no cover

    def on_available(self, peer_jid, stanza):
        """
        Callback called when a contact becomes available.
        To ve overloaded by user.

        Args:
          peer_jid (str): the JID of the agent that is available
          stanza (aioxmpp.Presence): The presence message containing type, show, priority and status values.

        """
        pass  # pragma: no cover

    def on_unavailable(self, peer_jid, stanza):
        """
        Callback called when a contact becomes unavailable.
        To ve overloaded by user.

        Args:
          peer_jid (str): the JID of the agent that is unavailable
          stanza (aioxmpp.Presence): The presence message containing type, show, priority and status values.

        """
        pass  # pragma: no cover
